# Copyright (C) 2021-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT
import json
import os
from abc import ABC, abstractmethod, abstractproperty, abstractstaticmethod
from contextlib import closing
from enum import Enum
from json.decoder import JSONDecodeError
from tempfile import NamedTemporaryFile

import av
from PIL import Image

from .utils import SortingMethod, md5_hash, rotate_image, sort


class VideoStreamReader:
    def __init__(self, source_path, chunk_size, force):
        self._source_path = source_path
        self._frames_number = None
        self._force = force
        self._upper_bound = 3 * chunk_size + 1

        with closing(av.open(self.source_path, mode="r")) as container:
            video_stream = VideoStreamReader._get_video_stream(container)
            isBreaked = False
            for packet in container.demux(video_stream):
                if isBreaked:
                    break
                for frame in packet.decode():
                    # check type of first frame
                    if not frame.pict_type.name == "I":
                        raise Exception("First frame is not key frame")

                    # get video resolution
                    if video_stream.metadata.get("rotate"):
                        frame = av.VideoFrame().from_ndarray(
                            rotate_image(
                                frame.to_ndarray(format="bgr24"),
                                360
                                - int(
                                    container.streams.video[0].metadata.get("rotate")
                                ),
                            ),
                            format="bgr24",
                        )
                    self.height, self.width = (frame.height, frame.width)
                    # not all videos contain information about numbers of frames
                    if video_stream.frames:
                        self._frames_number = video_stream.frames
                    isBreaked = True
                    break

    @property
    def source_path(self):
        return self._source_path

    @staticmethod
    def _get_video_stream(container):
        video_stream = next(
            stream for stream in container.streams if stream.type == "video"
        )
        video_stream.thread_type = "AUTO"
        return video_stream

    def __len__(self):
        return self._frames_number

    @property
    def resolution(self):
        return (self.width, self.height)

    def validate_key_frame(self, container, video_stream, key_frame):
        for packet in container.demux(video_stream):
            for frame in packet.decode():
                if md5_hash(frame) != key_frame["md5"] or frame.pts != key_frame["pts"]:
                    return False
                return True

    def __iter__(self):
        with closing(av.open(self.source_path, mode="r")) as container:
            video_stream = self._get_video_stream(container)
            frame_pts, frame_dts = -1, -1
            index, key_frame_number = 0, 0
            for packet in container.demux(video_stream):
                for frame in packet.decode():
                    if None not in {frame.pts, frame_pts} and frame.pts <= frame_pts:
                        raise Exception("Invalid pts sequences")
                    if None not in {frame.dts, frame_dts} and frame.dts <= frame_dts:
                        raise Exception("Invalid dts sequences")
                    frame_pts, frame_dts = frame.pts, frame.dts

                    if frame.key_frame:
                        key_frame_number += 1
                        ratio = (index + 1) // key_frame_number

                        if ratio >= self._upper_bound and not self._force:
                            raise AssertionError("Too few keyframes")

                        key_frame = {
                            "index": index,
                            "pts": frame.pts,
                            "md5": md5_hash(frame),
                        }

                        with closing(
                            av.open(self.source_path, mode="r")
                        ) as checked_container:
                            checked_container.seek(
                                offset=key_frame["pts"], stream=video_stream
                            )
                            isValid = self.validate_key_frame(
                                checked_container, video_stream, key_frame
                            )
                            if isValid:
                                yield (index, key_frame["pts"], key_frame["md5"])
                    else:
                        yield index
                    index += 1
            if not self._frames_number:
                self._frames_number = index


class KeyFramesVideoStreamReader(VideoStreamReader):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __iter__(self):
        with closing(av.open(self.source_path, mode="r")) as container:
            video_stream = self._get_video_stream(container)
            frame_pts, frame_dts = -1, -1
            index, key_frame_number = 0, 0
            for packet in container.demux(video_stream):
                for frame in packet.decode():
                    if None not in {frame.pts, frame_pts} and frame.pts <= frame_pts:
                        raise Exception("Invalid pts sequences")
                    if None not in {frame.dts, frame_dts} and frame.dts <= frame_dts:
                        raise Exception("Invalid dts sequences")
                    frame_pts, frame_dts = frame.pts, frame.dts

                    if frame.key_frame:
                        key_frame_number += 1
                        ratio = (index + 1) // key_frame_number
                        if ratio >= self._upper_bound and not self._force:
                            raise AssertionError("Too few keyframes")
                        key_frame = {
                            "index": index,
                            "pts": frame.pts,
                            "md5": md5_hash(frame),
                        }

                        with closing(
                            av.open(self.source_path, mode="r")
                        ) as checked_container:
                            checked_container.seek(
                                offset=key_frame["pts"], stream=video_stream
                            )
                            isValid = self.validate_key_frame(
                                checked_container, video_stream, key_frame
                            )
                            if isValid:
                                yield (index, key_frame["pts"], key_frame["md5"])
                    index += 1


class DatasetImagesReader:
    def __init__(
        self,
        sources,
        meta=None,
        sorting_method=SortingMethod.PREDEFINED,
        use_image_hash=False,
        start=0,
        step=1,
        stop=None,
        *args,
        **kwargs,
    ):
        self._sources = sort(sources, sorting_method)
        self._meta = meta
        self._data_dir = kwargs.get("data_dir", None)
        self._use_image_hash = use_image_hash
        self._start = start
        self._stop = stop if stop else len(sources)
        self._step = step

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value):
        self._start = int(value)

    @property
    def stop(self):
        return self._stop

    @stop.setter
    def stop(self, value):
        self._stop = int(value)

    @property
    def step(self):
        return self._step

    @step.setter
    def step(self, value):
        self._step = int(value)

    def __iter__(self):
        sources = (i for i in self._sources)
        for idx in range(self._stop):
            if idx in self.range_:
                image = next(sources)
                img = Image.open(image, mode="r")
                orientation = img.getexif().get(274, 1)
                img_name = (
                    os.path.relpath(image, self._data_dir)
                    if self._data_dir
                    else os.path.basename(image)
                )
                name, extension = os.path.splitext(img_name)
                width, height = img.width, img.height
                if orientation > 4:
                    width, height = height, width
                image_properties = {
                    "name": name.replace("\\", "/"),
                    "extension": extension,
                    "width": width,
                    "height": height,
                }
                if self._meta and img_name in self._meta:
                    image_properties["meta"] = self._meta[img_name]
                if self._use_image_hash:
                    image_properties["checksum"] = md5_hash(img)
                yield image_properties
            else:
                yield dict()

    @property
    def range_(self):
        return range(self._start, self._stop, self._step)

    def __len__(self):
        return len(self.range_)


class Dataset3DImagesReader(DatasetImagesReader):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __iter__(self):
        sources = (i for i in self._sources)
        for idx in range(self._stop):
            if idx in self.range_:
                image = next(sources)
                img_name = (
                    os.path.relpath(image, self._data_dir)
                    if self._data_dir
                    else os.path.basename(image)
                )
                name, extension = os.path.splitext(img_name)
                image_properties = {
                    "name": name,
                    "extension": extension,
                }
                if self._meta and img_name in self._meta:
                    image_properties["meta"] = self._meta[img_name]
                yield image_properties
            else:
                yield dict()


class _Manifest:
    class SupportedVersion(str, Enum):
        V1 = "1.0"
        V1_1 = "1.1"

        @classmethod
        def choices(cls):
            return (x.value for x in cls)

        def __str__(self):
            return self.value

    FILE_NAME = "manifest.jsonl"
    VERSION = SupportedVersion.V1_1

    def __init__(self, path, upload_dir=None):
        assert path, "A path to manifest file not found"
        self._path = os.path.join(path, self.FILE_NAME) if os.path.isdir(path) else path
        self._upload_dir = upload_dir

    @property
    def path(self):
        return self._path

    @property
    def name(self):
        return (
            os.path.basename(self._path)
            if not self._upload_dir
            else os.path.relpath(self._path, self._upload_dir)
        )


# Needed for faster iteration over the manifest file, will be generated to work inside CVAT
# and will not be generated when manually creating a manifest
class _Index:
    FILE_NAME = "index.json"

    def __init__(self, path):
        assert path and os.path.isdir(path), "No index directory path"
        self._path = os.path.join(path, self.FILE_NAME)
        self._index = {}

    @property
    def path(self):
        return self._path

    def dump(self):
        with open(self._path, "w") as index_file:
            json.dump(self._index, index_file, separators=(",", ":"))

    def load(self):
        with open(self._path, "r") as index_file:
            self._index = json.load(
                index_file, object_hook=lambda d: {int(k): v for k, v in d.items()}
            )

    def remove(self):
        os.remove(self._path)

    def create(self, manifest, skip):
        assert os.path.exists(
            manifest
        ), "A manifest file not exists, index cannot be created"
        with open(manifest, "r+") as manifest_file:
            while skip:
                manifest_file.readline()
                skip -= 1
            image_number = 0
            position = manifest_file.tell()
            line = manifest_file.readline()
            while line:
                if line.strip():
                    self._index[image_number] = position
                    image_number += 1
                    position = manifest_file.tell()
                line = manifest_file.readline()

    def partial_update(self, manifest, number):
        assert os.path.exists(
            manifest
        ), "A manifest file not exists, index cannot be updated"
        with open(manifest, "r+") as manifest_file:
            manifest_file.seek(self._index[number])
            line = manifest_file.readline()
            while line:
                if line.strip():
                    self._index[number] = manifest_file.tell()
                    number += 1
                line = manifest_file.readline()

    def __getitem__(self, number):
        assert 0 <= number < len(self), "Invalid index number: {}\nMax: {}".format(
            number, len(self) - 1
        )
        return self._index[number]

    def __len__(self):
        return len(self._index)


def _set_index(func):
    def wrapper(self, *args, **kwargs):
        func(self, *args, **kwargs)
        if self._create_index:
            self.set_index()

    return wrapper


class _ManifestManager(ABC):
    BASE_INFORMATION = {
        "version": 1,
        "type": 2,
    }

    def _json_item_is_valid(self, **state):
        for item in self._requared_item_attributes:
            if state.get(item, None) is None:
                raise Exception(
                    f"Invalid '{self.manifest.name} file structure': '{item}' is required, but not found"
                )

    def __init__(self, path, create_index, upload_dir=None, *args, **kwargs):
        self._manifest = _Manifest(path, upload_dir)
        self._index = _Index(os.path.dirname(self._manifest.path))
        self._reader = None
        self._create_index = create_index

    @property
    def reader(self):
        return self._reader

    def _parse_line(self, line):
        """Getting a random line from the manifest file"""
        with open(self._manifest.path, "r") as manifest_file:
            if isinstance(line, str):
                assert (
                    line in self.BASE_INFORMATION.keys()
                ), "An attempt to get non-existent information from the manifest"
                for _ in range(self.BASE_INFORMATION[line]):
                    fline = manifest_file.readline()
                return json.loads(fline)[line]
            else:
                assert self._index, "No prepared index"
                offset = self._index[line]
                manifest_file.seek(offset)
                properties = manifest_file.readline()
                parsed_properties = json.loads(properties)
                self._json_item_is_valid(**parsed_properties)
                return parsed_properties

    def init_index(self):
        if os.path.exists(self._index.path):
            self._index.load()
        else:
            self._index.create(
                self._manifest.path, 3 if self._manifest.TYPE == "video" else 2
            )
            self._index.dump()

    def reset_index(self):
        if os.path.exists(self._index.path):
            self._index.remove()

    def set_index(self):
        self.reset_index()
        self.init_index()

    def remove(self):
        self.reset_index()
        if os.path.exists(self.manifest.path):
            os.remove(self.manifest.path)

    @abstractmethod
    def create(self, content=None, _tqdm=None):
        pass

    @abstractmethod
    def partial_update(self, number, properties):
        pass

    def __iter__(self):
        with open(self._manifest.path, "r") as manifest_file:
            manifest_file.seek(self._index[0])
            image_number = 0
            line = manifest_file.readline()
            while line:
                if line.strip():
                    parsed_properties = json.loads(line)
                    self._json_item_is_valid(**parsed_properties)
                    yield (image_number, parsed_properties)
                    image_number += 1
                line = manifest_file.readline()

    @property
    def manifest(self):
        return self._manifest

    def __len__(self):
        if hasattr(self, "_index"):
            return len(self._index)
        else:
            return None

    def __getitem__(self, item):
        return self._parse_line(item)

    @property
    def index(self):
        return self._index

    @abstractproperty
    def data(self):
        pass

    @abstractmethod
    def get_subset(self, subset_names):
        pass


class VideoManifestManager(_ManifestManager):
    _requared_item_attributes = {"number", "pts"}

    def __init__(self, manifest_path, create_index=True):
        super().__init__(manifest_path, create_index)
        setattr(self._manifest, "TYPE", "video")
        self.BASE_INFORMATION["properties"] = 3

    def link(
        self,
        media_file,
        upload_dir=None,
        chunk_size=36,
        force=False,
        only_key_frames=False,
        **kwargs,
    ):
        ReaderClass = (
            VideoStreamReader if not only_key_frames else KeyFramesVideoStreamReader
        )
        self._reader = ReaderClass(
            os.path.join(upload_dir, media_file) if upload_dir else media_file,
            chunk_size,
            force,
        )

    def _write_base_information(self, file):
        base_info = {
            "version": self._manifest.VERSION,
            "type": self._manifest.TYPE,
            "properties": {
                "name": os.path.basename(self._reader.source_path),
                "resolution": self._reader.resolution,
                "length": len(self._reader),
            },
        }
        for key, value in base_info.items():
            json_item = json.dumps({key: value}, separators=(",", ":"))
            file.write(f"{json_item}\n")

    def _write_core_part(self, file, _tqdm):
        iterable_obj = (
            self._reader
            if _tqdm is None
            else _tqdm(self._reader, desc="Manifest creating", total=len(self._reader))
        )
        for item in iterable_obj:
            if isinstance(item, tuple):
                json_item = json.dumps(
                    {"number": item[0], "pts": item[1], "checksum": item[2]},
                    separators=(",", ":"),
                )
                file.write(f"{json_item}\n")

    # pylint: disable=arguments-differ
    @_set_index
    def create(self, _tqdm=None):
        """Creating and saving a manifest file"""
        if not len(self._reader):
            with NamedTemporaryFile(mode="w", delete=False) as tmp_file:
                self._write_core_part(tmp_file, _tqdm)
            temp = tmp_file.name
            with open(self._manifest.path, "w") as manifest_file:
                self._write_base_information(manifest_file)
                with open(temp, "r") as tmp_file:
                    manifest_file.write(tmp_file.read())
            os.remove(temp)
        else:
            with open(self._manifest.path, "w") as manifest_file:
                self._write_base_information(manifest_file)
                self._write_core_part(manifest_file, _tqdm)

    def partial_update(self, number, properties):
        pass

    @property
    def video_name(self):
        return self["properties"]["name"]

    @property
    def video_resolution(self):
        return self["properties"]["resolution"]

    @property
    def video_length(self):
        return self["properties"]["length"]

    @property
    def data(self):
        return self.video_name

    def get_subset(self, subset_names):
        raise NotImplementedError()


class VideoManifestValidator(VideoManifestManager):
    def __init__(self, source_path, manifest_path):
        self._source_path = source_path
        super().__init__(manifest_path)

    @staticmethod
    def _get_video_stream(container):
        video_stream = next(
            stream for stream in container.streams if stream.type == "video"
        )
        video_stream.thread_type = "AUTO"
        return video_stream

    def validate_key_frame(self, container, video_stream, key_frame):
        for packet in container.demux(video_stream):
            for frame in packet.decode():
                assert (
                    frame.pts == key_frame["pts"]
                ), "The uploaded manifest does not match the video"
                return

    def validate_seek_key_frames(self):
        with closing(av.open(self._source_path, mode="r")) as container:
            video_stream = self._get_video_stream(container)
            last_key_frame = None

            for _, key_frame in self:
                # check that key frames sequence sorted
                if last_key_frame and last_key_frame["number"] >= key_frame["number"]:
                    raise AssertionError(
                        "Invalid saved key frames sequence in manifest file"
                    )
                container.seek(offset=key_frame["pts"], stream=video_stream)
                self.validate_key_frame(container, video_stream, key_frame)
                last_key_frame = key_frame

    def validate_frame_numbers(self):
        with closing(av.open(self._source_path, mode="r")) as container:
            video_stream = self._get_video_stream(container)
            # not all videos contain information about numbers of frames
            frames = video_stream.frames
            if frames:
                assert (
                    frames == self.video_length
                ), "The uploaded manifest does not match the video"
                return


class ImageManifestManager(_ManifestManager):
    _requared_item_attributes = {"name", "extension"}

    def __init__(self, manifest_path, upload_dir=None, create_index=True):
        super().__init__(manifest_path, create_index, upload_dir)
        setattr(self._manifest, "TYPE", "images")

    def link(self, **kwargs):
        ReaderClass = (
            DatasetImagesReader
            if not kwargs.get("DIM_3D", None)
            else Dataset3DImagesReader
        )
        self._reader = ReaderClass(**kwargs)

    def _write_base_information(self, file):
        base_info = {
            "version": self._manifest.VERSION,
            "type": self._manifest.TYPE,
        }
        for key, value in base_info.items():
            json_line = json.dumps({key: value}, separators=(",", ":"))
            file.write(f"{json_line}\n")

    def _write_core_part(self, file, obj, _tqdm):
        iterable_obj = (
            obj
            if _tqdm is None
            else _tqdm(
                obj,
                desc="Manifest creating",
                total=None if not hasattr(obj, "__len__") else len(obj),
            )
        )
        for image_properties in iterable_obj:
            json_line = json.dumps(
                {key: value for key, value in image_properties.items()},
                separators=(",", ":"),
            )
            file.write(f"{json_line}\n")

    @_set_index
    def create(self, content=None, _tqdm=None):
        """Creating and saving a manifest file for the specialized dataset"""
        with open(self._manifest.path, "w") as manifest_file:
            self._write_base_information(manifest_file)
            obj = content if content else self._reader
            self._write_core_part(manifest_file, obj, _tqdm)

    def partial_update(self, number, properties):
        pass

    @property
    def data(self):
        return (f"{image['name']}{image['extension']}" for _, image in self)

    def get_subset(self, subset_names):
        index_list = []
        subset = []
        for _, image in self:
            image_name = f"{image['name']}{image['extension']}"
            if image_name in subset_names:
                index_list.append(subset_names.index(image_name))
                properties = {
                    "name": f"{image['name']}",
                    "extension": f"{image['extension']}",
                    "width": image["width"],
                    "height": image["height"],
                }
                for optional_field in {"meta", "checksum"}:
                    value = image.get(optional_field)
                    if value:
                        properties[optional_field] = value
                subset.append(properties)
        return index_list, subset


class _BaseManifestValidator(ABC):
    def __init__(self, full_manifest_path):
        self._manifest = _Manifest(full_manifest_path)

    def validate(self):
        try:
            # we cannot use index in general because manifest may be e.g. in share point with ro mode
            with open(self._manifest.path, "r") as manifest:
                for validator in self.validators:
                    line = json.loads(manifest.readline().strip())
                    validator(line)
            return True
        except (ValueError, KeyError, JSONDecodeError):
            return False

    @staticmethod
    def _validate_version(_dict):
        if not _dict["version"] in _Manifest.SupportedVersion.choices():
            raise ValueError("Incorrect version field")

    def _validate_type(self, _dict):
        if not _dict["type"] == self.TYPE:
            raise ValueError("Incorrect type field")

    @abstractproperty
    def validators(self):
        pass

    @abstractstaticmethod
    def _validate_first_item(_dict):
        pass


class _VideoManifestStructureValidator(_BaseManifestValidator):
    TYPE = "video"

    @property
    def validators(self):
        return (
            self._validate_version,
            self._validate_type,
            self._validate_properties,
            self._validate_first_item,
        )

    @staticmethod
    def _validate_properties(_dict):
        properties = _dict["properties"]
        if not isinstance(properties["name"], str):
            raise ValueError("Incorrect name field")
        if not isinstance(properties["resolution"], list):
            raise ValueError("Incorrect resolution field")
        if not isinstance(properties["length"], int) or properties["length"] == 0:
            raise ValueError("Incorrect length field")

    @staticmethod
    def _validate_first_item(_dict):
        if not isinstance(_dict["number"], int):
            raise ValueError("Incorrect number field")
        if not isinstance(_dict["pts"], int):
            raise ValueError("Incorrect pts field")


class _DatasetManifestStructureValidator(_BaseManifestValidator):
    TYPE = "images"

    @property
    def validators(self):
        return (
            self._validate_version,
            self._validate_type,
            self._validate_first_item,
        )

    @staticmethod
    def _validate_first_item(_dict):
        if not isinstance(_dict["name"], str):
            raise ValueError("Incorrect name field")
        if not isinstance(_dict["extension"], str):
            raise ValueError("Incorrect extension field")
        # FIXME
        # Width and height are required for 2D data, but
        # for 3D these parameters are not saved now.
        # It is necessary to uncomment these restrictions when manual preparation for 3D data is implemented.

        # if not isinstance(_dict['width'], int):
        #     raise ValueError('Incorrect width field')
        # if not isinstance(_dict['height'], int):
        #     raise ValueError('Incorrect height field')


def is_manifest(full_manifest_path):
    return _is_video_manifest(full_manifest_path) or _is_dataset_manifest(
        full_manifest_path
    )


def _is_video_manifest(full_manifest_path):
    validator = _VideoManifestStructureValidator(full_manifest_path)
    return validator.validate()


def _is_dataset_manifest(full_manifest_path):
    validator = _DatasetManifestStructureValidator(full_manifest_path)
    return validator.validate()



"""Tasks API calls."""
from __future__ import annotations

import math
import tempfile
import zipfile
from pathlib import Path
from typing import Any, Optional, Union

import cv2
from cvat_sdk.core.helpers import TqdmProgressReporter
from cvat_sdk.core.proxies.tasks import ResourceType
from tqdm import tqdm

from .auth import get_client
from .utils import VALID_IMAGE_FORMATS, VALID_VIDEO_FORMATS


def get_frames_info(task_id: int) -> list[dict[str, Any]]:
    """Get information about all frames in a task."""
    with get_client() as client:
        task = client.tasks.retrieve(task_id)
        return task.get_frames_info()


def create_task(name: str, project_id: int, segment_size: int = 25) -> dict[str, Any]:
    """Create a new task (without attached images/videos)."""
    data = {"name": name, "project_id": project_id, "segment_size": segment_size}
    with get_client() as client:
        task, _ = client.tasks.api.create(data)
    return task.to_dict()


def delete_task(task_id: int) -> None:
    """Delete a task."""
    with get_client() as client:
        task = client.tasks.retrieve(task_id)
        task.remove()


def list_tasks(project_id: int) -> list[int]:
    """List all tasks in a project."""
    with get_client() as client:
        return client.projects.retrieve(project_id).tasks


def download_annotations(
    task_id: int, output_path: Optional[Union[str, Path]] = None
) -> None:
    """Download CVAT XML annotations for a task."""
    if output_path is None:
        output_path = f"annotations_{task_id}.xml"
    with get_client() as client, tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = Path(tmp_dir)
        output_zip = tmp_dir / "download.zip"
        task = client.tasks.retrieve(task_id)
        task.export_dataset(
            "CVAT for images 1.1",
            output_zip,
            include_images=False,
            pbar=TqdmProgressReporter(tqdm()),
        )
        with zipfile.ZipFile(output_zip, "r") as zip:
            zip.extract("annotations.xml", tmp_dir)
        xml_path = tmp_dir / "annotations.xml"
        with open(xml_path, "r") as f1, open(output_path, "w") as f2:
            f2.write(f1.read())


def download_images(
    task_id: int, output_directory: Optional[Union[str, Path]] = None
) -> None:
    """Download CVAT images for a task."""
    if output_directory is None:
        output_directory = Path(f"images_{task_id}")
    output_directory = Path(output_directory)
    output_directory.mkdir(exist_ok=True, parents=True)
    with get_client() as client, tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = Path(tmp_dir)
        output_zip = tmp_dir / "download.zip"
        task = client.tasks.retrieve(task_id)
        print("Downloading images...")
        task.export_dataset(
            "CVAT for images 1.1",
            output_zip,
            include_images=True,
            pbar=TqdmProgressReporter(tqdm()),
        )
        with zipfile.ZipFile(output_zip, "r") as zip:
            zip.extractall(tmp_dir)
        print("Moving extracted images...")
        for image_path in tqdm(list((tmp_dir / "images").glob("*"))):
            with open(image_path, "rb") as f1, open(
                output_directory / image_path.name, "wb"
            ) as f2:
                f2.write(f1.read())


def upload_annotations(task_id: int, annotations_path: Union[str, Path]) -> None:
    """Upload annotations for a task."""
    annotations_path = str(annotations_path)
    with get_client() as client:
        task = client.tasks.retrieve(task_id)
        task.import_annotations(
            "CVAT 1.1", annotations_path, pbar=TqdmProgressReporter(tqdm())
        )


def upload_images(task_id: int, image_directory: Union[str, Path]) -> None:
    """Upload images from a directory to a task."""
    image_directory = Path(image_directory)
    images = []
    for extension in VALID_IMAGE_FORMATS:
        images.extend(image_directory.glob(f"*{extension}"))
    upload_image_list(task_id, images)


def upload_videos(project_id: int, video_directory: Union[str, Path]) -> None:
    """Upload video from a directory to a set of new tasks (1 per video)."""
    video_directory = Path(video_directory)
    videos: list[Path] = []
    for extension in VALID_VIDEO_FORMATS:
        videos.extend(video_directory.glob(f"*{extension}"))
    for video_path in videos:
        cap = cv2.VideoCapture(str(video_path))
        n_frames = math.ceil(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        task = create_task(video_path.name, project_id, segment_size=n_frames + 1)
        upload_image_list(task["id"], [video_path])


def upload_image_list(task_id: int, image_list: list[Union[str, Path]]) -> None:
    """Upload a list of images to a task."""
    with get_client() as client:
        task = client.tasks.retrieve(task_id)
        task.upload_data(
            ResourceType.LOCAL,
            [str(i) for i in image_list],
            pbar=TqdmProgressReporter(tqdm()),
        )



"""Expose CLI."""
import fire

from .annotations import boxes_to_cvat
from .dataset_manifest import create_manifest
from .jobs import complete_task, download_filename_urls, list_jobs
from .tasks import (
    create_task,
    delete_task,
    download_annotations,
    download_images,
    list_tasks,
    upload_annotations,
    upload_images,
    upload_videos,
)


def main() -> None:
    """Call CLI commands."""
    fire.Fire(
        {
            "boxes-to-cvat": boxes_to_cvat,
            "complete-task": complete_task,
            "delete-task": delete_task,
            "download-annotations": download_annotations,
            "download-filename-urls": download_filename_urls,
            "download-images": download_images,
            "create-task": create_task,
            "create-manifest": create_manifest,
            "list-jobs": list_jobs,
            "list-tasks": list_tasks,
            "upload-annotations": upload_annotations,
            "upload-images": upload_images,
            "upload-videos": upload_videos,
        }
    )



"""Jobs API calls."""
from __future__ import annotations

import os
from pathlib import Path
from typing import Optional, Union

import pandas as pd
from tqdm import tqdm

from .auth import get_client, get_ui_host


def list_jobs(task_id: int) -> list[int]:
    """List all tasks in a project."""
    with get_client() as client:
        return [job.id for job in client.tasks.retrieve(task_id).get_jobs()]


def update_stage(job_id: int, stage: str) -> None:
    """Update the stage of a job."""
    with get_client() as client:
        job = client.jobs.retrieve(job_id)
        job.update({"stage": stage})


def complete_task(task_id: int) -> None:
    """Complete a task."""
    job_ids = list_jobs(task_id)
    for job_id in tqdm(job_ids):
        update_stage(job_id, "acceptance")


def download_filename_urls(
    task_id: int, oputput_path: Optional[Union[str, Path]] = None
) -> None:
    """Create a CSV with the filename and URL for each image."""
    if oputput_path is None:
        oputput_path = f"filenames_{task_id}.csv"
    records = []
    host = get_ui_host()
    with get_client() as client:
        task = client.tasks.retrieve(task_id)
        for job in tqdm(task.get_jobs()):
            job_meta = job.get_meta()
            start_frame_id = job_meta["start_frame"]
            for i, frame in enumerate(job_meta["frames"]):
                route = f"tasks/{task_id}/jobs/{job.id}?frame={start_frame_id + i}"
                record = dict(filename=frame["name"], url=os.path.join(host, route))
                records.append(record)
    df = pd.DataFrame(records)
    df.to_csv(oputput_path, index=False)



# Copyright (C) 2021-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT
import hashlib
import mimetypes
import os
import re
from enum import Enum
from random import shuffle

import cv2 as cv
from av import VideoFrame
from natsort import os_sorted


def rotate_image(image, angle):
    height, width = image.shape[:2]
    image_center = (width / 2, height / 2)
    matrix = cv.getRotationMatrix2D(image_center, angle, 1.0)
    abs_cos = abs(matrix[0, 0])
    abs_sin = abs(matrix[0, 1])
    bound_w = int(height * abs_sin + width * abs_cos)
    bound_h = int(height * abs_cos + width * abs_sin)
    matrix[0, 2] += bound_w / 2 - image_center[0]
    matrix[1, 2] += bound_h / 2 - image_center[1]
    matrix = cv.warpAffine(image, matrix, (bound_w, bound_h))
    return matrix


def md5_hash(frame):
    if isinstance(frame, VideoFrame):
        frame = frame.to_image()
    return hashlib.md5(frame.tobytes()).hexdigest()  # nosec


def _define_data_type(media):
    return mimetypes.guess_type(media)[0]


def is_video(media_file):
    data_type = _define_data_type(media_file)
    return data_type is not None and data_type.startswith("video")


def is_image(media_file):
    data_type = _define_data_type(media_file)
    return (
        data_type is not None
        and data_type.startswith("image")
        and not data_type.startswith("image/svg")
    )


def _list_and_join(root):
    files = os.listdir(root)
    for f in files:
        yield os.path.join(root, f)


def _prepare_context_list(files, base_dir):
    return sorted(map(lambda x: os.path.relpath(x, base_dir), filter(is_image, files)))


# Expected 2D format is:
# data/
#   00001.png
#   related_images/
#     00001_png/
#       context_image_1.jpeg
#       context_image_2.png
def _detect_related_images_2D(image_paths, root_path):
    related_images = {}
    latest_dirname = ""
    related_images_exist = False

    for image_path in sorted(image_paths):
        rel_image_path = os.path.relpath(image_path, root_path)
        dirname = os.path.dirname(image_path)
        related_images_dirname = os.path.join(dirname, "related_images")
        related_images[rel_image_path] = []

        if latest_dirname == dirname and not related_images_exist:
            continue
        elif latest_dirname != dirname:
            # Update some data applicable for a subset of paths (within the current dirname)
            latest_dirname = dirname
            related_images_exist = os.path.isdir(related_images_dirname)

        if related_images_exist:
            related_images_dirname = os.path.join(
                related_images_dirname,
                "_".join(os.path.basename(image_path).rsplit(".", 1)),
            )

            if os.path.isdir(related_images_dirname):
                related_images[rel_image_path] = _prepare_context_list(
                    _list_and_join(related_images_dirname), root_path
                )
    return related_images


# Possible 3D formats are:
# velodyne_points/
#     data/
#         image_01.bin
# IMAGE_00 # any number?
#     data/
#         image_01.png

# pointcloud/
#     00001.pcd
# related_images/
#     00001_pcd/
#         image_01.png # or other image

# Default formats
# Option 1
# data/
#     image.pcd
#     image.png


# Option 2
# data/
#    image_1/
#        image_1.pcd
#        context_1.png
#        context_2.jpg
def _detect_related_images_3D(image_paths, root_path):
    related_images = {}
    latest_dirname = ""
    dirname_files = []
    related_images_exist = False
    velodyne_context_images_dirs = []

    for image_path in sorted(image_paths):
        rel_image_path = os.path.relpath(image_path, root_path)
        name = os.path.splitext(os.path.basename(image_path))[0]
        dirname = os.path.dirname(image_path)
        related_images_dirname = os.path.normpath(
            os.path.join(dirname, "..", "related_images")
        )
        related_images[rel_image_path] = []

        if latest_dirname != dirname:
            # Update some data applicable for a subset of paths (within the current dirname)
            latest_dirname = dirname
            related_images_exist = os.path.isdir(related_images_dirname)
            dirname_files = list(_list_and_join(dirname))
            velodyne_context_images_dirs = [
                directory
                for directory in _list_and_join(
                    os.path.normpath(os.path.join(dirname, "..", ".."))
                )
                if os.path.isdir(os.path.join(directory, "data"))
                and re.search(r"image_\d.*", directory, re.IGNORECASE)
            ]

        filtered_dirname_files = list(filter(lambda x: x != image_path, dirname_files))
        if len(filtered_dirname_files) and os.path.basename(dirname) == name:
            # default format (option 2)
            related_images[rel_image_path].extend(
                _prepare_context_list(filtered_dirname_files, root_path)
            )
        else:
            filtered_dirname_files = list(
                filter(
                    lambda x: os.path.splitext(os.path.basename(x))[0] == name,
                    filtered_dirname_files,
                )
            )
            if len(filtered_dirname_files):
                # default format (option 1)
                related_images[rel_image_path].extend(
                    _prepare_context_list(filtered_dirname_files, root_path)
                )

        if related_images_exist:
            related_images_dirname = os.path.join(
                related_images_dirname,
                "_".join(os.path.basename(image_path).rsplit(".", 1)),
            )
            if os.path.isdir(related_images_dirname):
                related_images[rel_image_path].extend(
                    _prepare_context_list(
                        _list_and_join(related_images_dirname), root_path
                    )
                )

        if dirname.endswith(os.path.join("velodyne_points", "data")):
            # velodynepoints format
            for context_images_dir in velodyne_context_images_dirs:
                context_files = _list_and_join(os.path.join(context_images_dir, "data"))
                context_files = list(
                    filter(
                        lambda x: os.path.splitext(os.path.basename(x))[0] == name,
                        context_files,
                    )
                )
                related_images[rel_image_path].extend(
                    _prepare_context_list(context_files, root_path)
                )

        related_images[rel_image_path].sort()
    return related_images


# This function is expected to be called only for images tasks
# image_path is expected to be a list of absolute path to images
# root_path is expected to be a string (dataset root)
def detect_related_images(image_paths, root_path):
    data_are_2d = False
    data_are_3d = False

    # First of all need to define data type we are working with
    for image_path in image_paths:
        # .bin files are expected to be converted to .pcd before this code
        if os.path.splitext(image_path)[1].lower() == ".pcd":
            data_are_3d = True
        else:
            data_are_2d = True
    assert not (
        data_are_3d and data_are_2d
    ), "Combined data types 2D and 3D are not supported"

    if data_are_2d:
        return _detect_related_images_2D(image_paths, root_path)
    elif data_are_3d:
        return _detect_related_images_3D(image_paths, root_path)
    return {}


class SortingMethod(str, Enum):
    LEXICOGRAPHICAL = "lexicographical"
    NATURAL = "natural"
    PREDEFINED = "predefined"
    RANDOM = "random"

    @classmethod
    def choices(cls):
        return tuple((x.value, x.name) for x in cls)

    def __str__(self):
        return self.value


def sort(images, sorting_method=SortingMethod.LEXICOGRAPHICAL, func=None):
    if sorting_method == SortingMethod.LEXICOGRAPHICAL:
        return sorted(images, key=func)
    elif sorting_method == SortingMethod.NATURAL:
        return os_sorted(images, key=func)
    elif sorting_method == SortingMethod.PREDEFINED:
        return images
    elif sorting_method == SortingMethod.RANDOM:
        shuffle(images)
        return images
    else:
        raise NotImplementedError()



"""Projects API calls."""
from __future__ import annotations

from typing import Any

from .auth import get_client


def get_frames_info(project_id: int) -> list[dict[str, Any]]:
    """Get information about all frames in a task."""
    with get_client() as client:
        frames_info = []
        for task in client.projects.retrieve(project_id).get_tasks():
            frames_info.extend(task.get_frames_info())
        return frames_info



# Copyright (C) 2021-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT
import os
import re
import sys
from glob import glob

from tqdm import tqdm

from .core import ImageManifestManager, VideoManifestManager
from .utils import detect_related_images, is_image, is_video


def create_manifest(
    source: str,
    output_dir: str = os.getcwd(),
    sorting: str = "lexicographical",
    force: bool = False,
) -> None:
    """
    Create a manifest for a set of images or videos.

    Parameters
    ----------
    source
        Source paths
    output_dir
        Directory where the manifest file will be saved
    sorting
        One of "lexicographical", "natural", "predefined", "random"
    """
    manifest_directory = os.path.abspath(output_dir)
    if not os.path.exists(manifest_directory):
        os.makedirs(manifest_directory)
    source = os.path.abspath(os.path.expanduser(source))

    sources = []
    if not os.path.isfile(source):  # directory/pattern with images
        data_dir = None
        if os.path.isdir(source):
            data_dir = source
            for root, _, files in os.walk(source):
                sources.extend([os.path.join(root, f) for f in files if is_image(f)])
        else:
            items = source.lstrip("/").split("/")
            position = 0
            try:
                for item in items:
                    if set(item) & {"*", "?", "[", "]"}:
                        break
                    position += 1
                else:
                    raise Exception("Wrong positional argument")
                assert position != 0, "Wrong pattern: there must be a common root"
                data_dir = source.split(items[position])[0]
            except Exception as ex:
                sys.exit(str(ex))
            sources = list(filter(is_image, glob(source, recursive=True)))

        sources = list(
            filter(lambda x: "related_images{}".format(os.sep) not in x, sources)
        )

        # If the source is a glob expression, we need additional processing
        abs_root = source
        while abs_root and re.search(r"[*?\[\]]", abs_root):
            abs_root = os.path.split(abs_root)[0]

        related_images = detect_related_images(sources, abs_root)
        meta = {k: {"related_images": related_images[k]} for k in related_images}
        try:
            assert len(sources), "A images was not found"
            manifest = ImageManifestManager(manifest_path=manifest_directory)
            manifest.link(
                sources=sources,
                meta=meta,
                sorting_method=sorting,
                use_image_hash=True,
                data_dir=data_dir,
            )
            manifest.create(_tqdm=tqdm)
        except Exception as ex:
            sys.exit(str(ex))
    else:  # video
        try:
            assert is_video(
                source
            ), "You can specify a video path or a directory/pattern with images"
            manifest = VideoManifestManager(manifest_path=manifest_directory)
            manifest.link(media_file=source, force=force)
            try:
                manifest.create(_tqdm=tqdm)
            except AssertionError as ex:
                if str(ex) == "Too few keyframes":
                    msg = (
                        "NOTE: prepared manifest file contains too few key frames for smooth decoding.\n"
                        "Use --force flag if you still want to prepare a manifest file."
                    )
                    print(msg)
                    sys.exit(2)
                else:
                    raise
        except Exception as ex:
            sys.exit(str(ex))

    print("The manifest file has been prepared")



"""Convert to CVAT XML."""
from __future__ import annotations

import tempfile
from pathlib import Path
from typing import Union

import datumaro as dm
import numpy as np
from sparrow_datums import FrameAugmentedBoxes


def boxes_to_cvat(
    annotation_directory: Union[str, Path],
    output_path: Union[str, Path],
    labels: list[str],
) -> None:
    """Convert a directory of FrameAugmentedBoxes to CVAT XML."""
    annotation_directory = Path(annotation_directory)
    output_path = Path(output_path)
    dataset_items = []
    for annotation_path in annotation_directory.glob("*.json.gz"):
        boxes = FrameAugmentedBoxes.from_file(annotation_path)
        annotations = []
        for box in boxes.to_absolute():
            annotations.append(dm.Bbox(box.x, box.y, box.w, box.h, label=box.label))
        dataset_item = dm.DatasetItem(
            id=annotation_path.name.rstrip(".json.gz"),
            media=dm.Image(np.empty((boxes.image_height, boxes.image_width, 3))),
            annotations=annotations,
        )
        dataset_items.append(dataset_item)
    dataset = dm.Dataset.from_iterable(dataset_items, categories=list(labels))
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = Path(tmp_dir)
        dataset.export(tmp_dir, "cvat")
        with open(tmp_dir / "default.xml", "r") as f1, open(output_path, "w") as f2:
            f2.write(f1.read())
